package sun.reflect.annotation;

import java.lang.reflect.AnnotatedWildcardType;
import java.util.ArrayList;
import java.util.Arrays;
import java.lang.reflect.AnnotatedParameterizedType;
import java.lang.reflect.AnnotatedTypeVariable;
import java.lang.reflect.AnnotatedArrayType;
import java.lang.annotation.Annotation;
import java.util.Map;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.Modifier;
import java.lang.reflect.WildcardType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.AnnotatedElement;
import java.lang.reflect.Type;
import java.lang.reflect.AnnotatedType;

public final class AnnotatedTypeFactory
{
    static final AnnotatedType EMPTY_ANNOTATED_TYPE;
    static final AnnotatedType[] EMPTY_ANNOTATED_TYPE_ARRAY;
    
    public AnnotatedTypeFactory() {
    }
    
    public static AnnotatedType buildAnnotatedType(Type obj, TypeAnnotation.LocationInfo locationInfo, TypeAnnotation[] array, TypeAnnotation[] array2, AnnotatedElement annotatedElement) {
        if (obj == null) {
            return AnnotatedTypeFactory.EMPTY_ANNOTATED_TYPE;
        }
        if (isArray(obj)) {
            return new AnnotatedTypeFactory.AnnotatedArrayTypeImpl(obj, locationInfo, array, array2, annotatedElement);
        }
        if ((obj instanceof Class)) {
            return new AnnotatedTypeFactory.AnnotatedTypeBaseImpl(obj, addNesting(obj, locationInfo), array, array2, annotatedElement);
        }
        if ((obj instanceof TypeVariable)) {
            return new AnnotatedTypeFactory.AnnotatedTypeVariableImpl((TypeVariable)obj, locationInfo, array, array2, annotatedElement);
        }
        if ((obj instanceof ParameterizedType)) {
            return new AnnotatedTypeFactory.AnnotatedParameterizedTypeImpl((ParameterizedType)obj, addNesting(obj, locationInfo), array, array2, annotatedElement);
        }
        if ((obj instanceof WildcardType)) {
            return new AnnotatedTypeFactory.AnnotatedWildcardTypeImpl((WildcardType)obj, locationInfo, array, array2, annotatedElement);
        }
        throw new AssertionError("Unknown instance of Type: " + obj + "\nThis should not happen.");
    }
    
    private static TypeAnnotation.LocationInfo addNesting(Type type, TypeAnnotation.LocationInfo locationInfo) {
        if (isArray(type)) {
            return locationInfo;
        }
        if ((type instanceof Class)) {
            Class<Object> clazz = (Class)type;
            if (clazz.getEnclosingClass() == null) {
                return locationInfo;
            }
            if (Modifier.isStatic(clazz.getModifiers())) {
                return addNesting(clazz.getEnclosingClass(), locationInfo);
            }
            return addNesting(clazz.getEnclosingClass(), locationInfo.pushInner());
        }
        else if ((type instanceof ParameterizedType)) {
            ParameterizedType parameterizedType = (ParameterizedType)type;
            if (parameterizedType.getOwnerType() == null) {
                return locationInfo;
            }
            return addNesting(parameterizedType.getOwnerType(), locationInfo.pushInner());
        }
        else {
            return locationInfo;
        }
    }
    
    private static boolean isArray(Type type) {
        if ((type instanceof Class)) {
            if (((Class)type).isArray()) {
                return true;
            }
        }
        else if ((type instanceof GenericArrayType)) {
            return true;
        }
        return false;
    }
    
    static {
        EMPTY_ANNOTATED_TYPE = new AnnotatedTypeFactory.AnnotatedTypeBaseImpl(null, TypeAnnotation.LocationInfo.BASE_LOCATION, new TypeAnnotation[0], new TypeAnnotation[0], null);
        EMPTY_ANNOTATED_TYPE_ARRAY = new AnnotatedType[0];
    }
    
    private static class AnnotatedTypeBaseImpl implements AnnotatedType
    {
        private final Type type;
        private final AnnotatedElement decl;
        private final TypeAnnotation.LocationInfo location;
        private final TypeAnnotation[] allOnSameTargetTypeAnnotations;
        private final Map<Class<? extends Annotation>, Annotation> annotations;
        
        AnnotatedTypeBaseImpl(Type type, TypeAnnotation.LocationInfo location, TypeAnnotation[] array, TypeAnnotation[] allOnSameTargetTypeAnnotations, AnnotatedElement decl) {
            this.type = type;
            this.decl = decl;
            this.location = location;
            this.allOnSameTargetTypeAnnotations = allOnSameTargetTypeAnnotations;
            this.annotations = TypeAnnotationParser.mapTypeAnnotations(location.filter(array));
        }
        
        @Override
        public final Annotation[] getAnnotations() {
            return this.getDeclaredAnnotations();
        }
        
        @Override
        public final <T extends Annotation> T getAnnotation(Class<T> clazz) {
            return this.getDeclaredAnnotation(clazz);
        }
        
        @Override
        public final <T extends Annotation> T[] getAnnotationsByType(Class<T> clazz) {
            return this.getDeclaredAnnotationsByType(clazz);
        }
        
        @Override
        public final Annotation[] getDeclaredAnnotations() {
            return (Annotation[])this.annotations.values().toArray(new Annotation[0]);
        }
        
        @Override
        public final <T extends Annotation> T getDeclaredAnnotation(Class<T> clazz) {
            return (T)this.annotations.get(clazz);
        }
        
        @Override
        public final <T extends Annotation> T[] getDeclaredAnnotationsByType(Class<T> clazz) {
            return AnnotationSupport.getDirectlyAndIndirectlyPresent(this.annotations, clazz);
        }
        
        @Override
        public final Type getType() {
            return this.type;
        }
        
        final TypeAnnotation.LocationInfo getLocation() {
            return this.location;
        }
        
        final TypeAnnotation[] getTypeAnnotations() {
            return this.allOnSameTargetTypeAnnotations;
        }
        
        final AnnotatedElement getDecl() {
            return this.decl;
        }
    }
    
    private static final class AnnotatedArrayTypeImpl extends AnnotatedTypeFactory.AnnotatedTypeBaseImpl implements AnnotatedArrayType
    {
        AnnotatedArrayTypeImpl(Type type, TypeAnnotation.LocationInfo locationInfo, TypeAnnotation[] array, TypeAnnotation[] array2, AnnotatedElement annotatedElement) {
            super(type, locationInfo, array, array2, annotatedElement);
        }
        
        @Override
        public AnnotatedType getAnnotatedGenericComponentType() {
            return AnnotatedTypeFactory.buildAnnotatedType(this.getComponentType(), this.getLocation().pushArray(), this.getTypeAnnotations(), this.getTypeAnnotations(), this.getDecl());
        }
        
        private Type getComponentType() {
            Type type = this.getType();
            if ((type instanceof Class)) {
                return ((Class)type).getComponentType();
            }
            return ((GenericArrayType)type).getGenericComponentType();
        }
    }
    
    private static final class AnnotatedTypeVariableImpl extends AnnotatedTypeFactory.AnnotatedTypeBaseImpl implements AnnotatedTypeVariable
    {
        AnnotatedTypeVariableImpl(TypeVariable<?> typeVariable, TypeAnnotation.LocationInfo locationInfo, TypeAnnotation[] array, TypeAnnotation[] array2, AnnotatedElement annotatedElement) {
            super(typeVariable, locationInfo, array, array2, annotatedElement);
        }
        
        @Override
        public AnnotatedType[] getAnnotatedBounds() {
            return this.getTypeVariable().getAnnotatedBounds();
        }
        
        private TypeVariable<?> getTypeVariable() {
            return (TypeVariable)this.getType();
        }
    }
    
    private static final class AnnotatedParameterizedTypeImpl extends AnnotatedTypeFactory.AnnotatedTypeBaseImpl implements AnnotatedParameterizedType
    {
        AnnotatedParameterizedTypeImpl(ParameterizedType parameterizedType, TypeAnnotation.LocationInfo locationInfo, TypeAnnotation[] array, TypeAnnotation[] array2, AnnotatedElement annotatedElement) {
            super(parameterizedType, locationInfo, array, array2, annotatedElement);
        }
        
        @Override
        public AnnotatedType[] getAnnotatedActualTypeArguments() {
            Type[] actualTypeArguments = this.getParameterizedType().getActualTypeArguments();
            AnnotatedType[] a = new AnnotatedType[actualTypeArguments.length];
            Arrays.fill(a, AnnotatedTypeFactory.EMPTY_ANNOTATED_TYPE);
            int length = this.getTypeAnnotations().length;
            Object o;
            TypeAnnotation.LocationInfo pushTypeArg;
            for (int i = 0; i < a.length; a[i] = AnnotatedTypeFactory.buildAnnotatedType(actualTypeArguments[i], pushTypeArg, (TypeAnnotation[])o.toArray(new TypeAnnotation[0]), this.getTypeAnnotations(), this.getDecl()), ++i) {
                o = new ArrayList(length);
                pushTypeArg = this.getLocation().pushTypeArg((short)(byte)i);
                for (TypeAnnotation typeAnnotation : this.getTypeAnnotations()) {
                    if (typeAnnotation.getLocationInfo().isSameLocationInfo(pushTypeArg)) {
                        o.add(typeAnnotation);
                    }
                }
            }
            return a;
        }
        
        private ParameterizedType getParameterizedType() {
            return (ParameterizedType)this.getType();
        }
    }
    
    private static final class AnnotatedWildcardTypeImpl extends AnnotatedTypeFactory.AnnotatedTypeBaseImpl implements AnnotatedWildcardType
    {
        private final boolean hasUpperBounds;
        
        AnnotatedWildcardTypeImpl(WildcardType wildcardType, TypeAnnotation.LocationInfo locationInfo, TypeAnnotation[] array, TypeAnnotation[] array2, AnnotatedElement annotatedElement) {
            super(wildcardType, locationInfo, array, array2, annotatedElement);
            this.hasUpperBounds = (wildcardType.getLowerBounds().length == 0);
        }
        
        @Override
        public AnnotatedType[] getAnnotatedUpperBounds() {
            if (!this.hasUpperBounds()) {
                return new AnnotatedType[0];
            }
            return this.getAnnotatedBounds(this.getWildcardType().getUpperBounds());
        }
        
        @Override
        public AnnotatedType[] getAnnotatedLowerBounds() {
            if (this.hasUpperBounds) {
                return new AnnotatedType[0];
            }
            return this.getAnnotatedBounds(this.getWildcardType().getLowerBounds());
        }
        
        private AnnotatedType[] getAnnotatedBounds(Type[] array) {
            AnnotatedType[] a = new AnnotatedType[array.length];
            Arrays.fill(a, AnnotatedTypeFactory.EMPTY_ANNOTATED_TYPE);
            TypeAnnotation.LocationInfo pushWildcard = this.getLocation().pushWildcard();
            int length = this.getTypeAnnotations().length;
            Object o;
            for (int i = 0; i < a.length; a[i] = AnnotatedTypeFactory.buildAnnotatedType(array[i], pushWildcard, (TypeAnnotation[])o.toArray(new TypeAnnotation[0]), this.getTypeAnnotations(), this.getDecl()), ++i) {
                o = new ArrayList(length);
                for (TypeAnnotation typeAnnotation : this.getTypeAnnotations()) {
                    if (typeAnnotation.getLocationInfo().isSameLocationInfo(pushWildcard)) {
                        o.add(typeAnnotation);
                    }
                }
            }
            return a;
        }
        
        private WildcardType getWildcardType() {
            return (WildcardType)this.getType();
        }
        
        private boolean hasUpperBounds() {
            return this.hasUpperBounds;
        }
    }
}
